==> lightphe-0.0.4.tar.gz: lightphe-0.0.4/lightphe/cryptosystems/EllipticCurveElGamal.py <==
import random
from typing import Optional
from lightphe.models.Homomorphic import Homomorphic
from lightphe.elliptic.Weierstrass import Weierstrass
from lightphe.commons.logger import Logger

logger = Logger()


class EllipticCurveElGamal(Homomorphic):
    """
    Elliptic Curve ElGamal algorithm is an additively homomorphic algorithm
    Unluckily, it requires to solve (EC)DLP to restore plaintext in decryption
    However it is easy to restore plaintext while plaintext is not very large
    unsimilar to Benaloh or Naccache-Stern
    Ref: https://sefiks.com/2018/08/21/elliptic-curve-elgamal-encryption/
    """

    def __init__(self, keys: Optional[dict] = None, key_size: int = 160):
        """
        Args:
            keys (dict): private - public key pair.
                set this to None if you want to generate random keys.
            key_size (int): key size in bits. default is 160.
                this is equivalent to 1024 bit RSA.
        """
        # TODO: add different forms and curves. e.g. Koblitz, Edwards (Ed25519)
        self.curve = Weierstrass()
        self.keys = keys or self.generate_keys(key_size)
        self.modulo = self.curve.p

    def generate_keys(self, key_size: int):
        """
        Generate public and private keys of Elliptic Curve ElGamal cryptosystem
        Args:
            key_size (int): key size in bits
        Returns:
            keys (dict): having private_key and public_key keys
        """
        keys = {}
        keys["private_key"] = {}
        keys["public_key"] = {}

        # private key
        ka = random.getrandbits(key_size)

        # public key
        Qa = self.curve.apply_double_and_add_method(G=self.curve.G, k=ka, p=self.curve.p)

        keys["public_key"]["Qa"] = Qa
        keys["private_key"]["ka"] = ka

        return keys

    def generate_random_key(self) -> int:
        """
        Elliptic Curve ElGamal requires to generate one-time random key per encryption
        Returns:
            random key (int): one time random key for encryption
        """
        return random.getrandbits(128)

    def encrypt(self, plaintext: int, random_key: Optional[int] = None) -> tuple:
        """
        Encrypt plaintext with Elliptic Curve ElGamal
        Args:
            plaintext (int): message to encrypt
            random_key (int): random key for encryption. Do not set this to a static value.
        Returns
            ciphertext (tuple): c1 and c2
        """
        # modulo
        p = self.curve.p

        # base point
        G = self.curve.G

        # public key
        Qa = self.keys["public_key"]["Qa"]

        # random key
        r = random_key or self.generate_random_key()

        s = self.curve.apply_double_and_add_method(G=G, k=plaintext, p=p)

        c1 = self.curve.apply_double_and_add_method(G=G, k=r, p=p)

        c2 = self.curve.apply_double_and_add_method(G=Qa, k=r, p=p)
        c2 = self.curve.add_points(c2, s, p)

        return c1, c2

    def decrypt(self, ciphertext: tuple) -> int:
        """
        Decrypt ciphertext with Elliptic Curve ElGamal
        Args:
            ciphertext (tuple): c1 and c2
        Returns:
            plaintext (int): restored message
        """
        # modulo
        p = self.curve.p

        # private key
        ka = self.keys["private_key"]["ka"]

        c1, c2 = ciphertext
        c1_prime = (c1[0], (-1 * c1[1]) % p)
        s_prime = self.curve.apply_double_and_add_method(G=c1_prime, k=ka, p=p)
        s_prime = self.curve.add_points(P=c2, Q=s_prime, p=p)

        # s_prime is a point on the elliptic curve
        # s_prime = k x G
        # we need to find k from known s_prime and G
        # this requires to solve ECDLP

        # base point
        G = self.curve.G
        k = 2
        while True:
            G = self.curve.add_points(P=G, Q=self.curve.G, p=p)
            if G[0] == s_prime[0] and G[1] == s_prime[1]:
                return k
            k = k + 1
            if k > self.curve.n:
                raise ValueError(f"Cannot restore scalar from {s_prime} = k x {self.curve.G}")

    def multiply(self, ciphertext1: tuple, ciphertext2: tuple) -> tuple:
        raise ValueError(
            "Elliptic Curve ElGamal is not homomorphic with respect to the multiplication"
        )

    def add(self, ciphertext1: tuple, ciphertext2: tuple) -> tuple:
        """
        Perform homomorphic addition on encrypted data
        Result of this must be equal to E(m1 + m2)
        Args:
            ciphertext1 (dict): Elliptic Curve ElGamal ciphertext consisting of c1 and c2 keys
            ciphertext2 (dict): Elliptic Curve ElGamal ciphertext consisting of c1 and c2 keys
        Returns
            ciphertext (dict): Elliptic Curve ElGamal ciphertext consisting of c1 and c2 keys
        """
        a = self.curve.add_points(P=ciphertext1[0], Q=ciphertext2[0], p=self.curve.p)
        b = self.curve.add_points(P=ciphertext1[1], Q=ciphertext2[1], p=self.curve.p)
        return a, b

    def xor(self, ciphertext1: tuple, ciphertext2: tuple) -> int:
        raise ValueError(
            "Elliptic Curve ElGamal is not homomorphic with respect to the exclusive or"
        )

    def multiply_by_contant(self, ciphertext: tuple, constant: int) -> tuple:
        """
        Multiply a ciphertext with a plain constant.
        Result of this must be equal to k x E(m1) = E(m1 * k)
        where E(m1) = ciphertext
        Args:
            ciphertext (int): ciphertext created with Elliptic Curve ElGamal
            constant (int): known plain constant
        Returns:
            ciphertext (int): new ciphertext created with Elliptic Curve ElGamal
        """
        return self.curve.apply_double_and_add_method(
            G=ciphertext[0], k=constant, p=self.curve.p
        ), self.curve.apply_double_and_add_method(G=ciphertext[1], k=constant, p=self.curve.p)

    def reencrypt(self, ciphertext: tuple) -> tuple:
        raise ValueError("Elliptic Curve ElGamal does not support regeneration of ciphertext")

==> lightphe-0.0.7.tar.gz: lightphe-0.0.7/lightphe/cryptosystems/EllipticCurveElGamal.py <==
import random
from typing import Optional
from lightphe.models.Homomorphic import Homomorphic
from lightphe.elliptic.Weierstrass import Weierstrass
from lightphe.commons.logger import Logger

logger = Logger(module="lightphe/cryptosystems/EllipticCurveElGamal.py")


class EllipticCurveElGamal(Homomorphic):
    """
    Elliptic Curve ElGamal algorithm is an additively homomorphic algorithm
    Unluckily, it requires to solve (EC)DLP to restore plaintext in decryption
    However it is easy to restore plaintext while plaintext is not very large
    unsimilar to Benaloh or Naccache-Stern
    Ref: https://sefiks.com/2018/08/21/elliptic-curve-elgamal-encryption/
    """

    def __init__(self, keys: Optional[dict] = None, key_size: int = 160):
        """
        Args:
            keys (dict): private - public key pair.
                set this to None if you want to generate random keys.
            key_size (int): key size in bits. default is 160.
                this is equivalent to 1024 bit RSA.
        """
        # TODO: add different forms and curves. e.g. Koblitz, Edwards (Ed25519)
        self.curve = Weierstrass()
        self.keys = keys or self.generate_keys(key_size)
        self.plaintext_modulo = self.curve.p
        self.ciphertext_modulo = self.curve.p

    def generate_keys(self, key_size: int):
        """
        Generate public and private keys of Elliptic Curve ElGamal cryptosystem
        Args:
            key_size (int): key size in bits
        Returns:
            keys (dict): having private_key and public_key keys
        """
        keys = {}
        keys["private_key"] = {}
        keys["public_key"] = {}

        # private key
        ka = random.getrandbits(key_size)

        # public key
        Qa = self.curve.apply_double_and_add_method(G=self.curve.G, k=ka, p=self.curve.p)

        keys["public_key"]["Qa"] = Qa
        keys["private_key"]["ka"] = ka

        return keys

    def generate_random_key(self) -> int:
        """
        Elliptic Curve ElGamal requires to generate one-time random key per encryption
        Returns:
            random key (int): one time random key for encryption
        """
        return random.getrandbits(128)

    def encrypt(self, plaintext: int, random_key: Optional[int] = None) -> tuple:
        """
        Encrypt plaintext with Elliptic Curve ElGamal
        Args:
            plaintext (int): message to encrypt
            random_key (int): random key for encryption. Do not set this to a static value.
        Returns
            ciphertext (tuple): c1 and c2
        """
        # modulo
        p = self.curve.p

        # base point
        G = self.curve.G

        # public key
        Qa = self.keys["public_key"]["Qa"]

        # random key
        r = random_key or self.generate_random_key()

        s = self.curve.apply_double_and_add_method(G=G, k=plaintext, p=p)

        c1 = self.curve.apply_double_and_add_method(G=G, k=r, p=p)

        c2 = self.curve.apply_double_and_add_method(G=Qa, k=r, p=p)
        c2 = self.curve.add_points(c2, s, p)

        return c1, c2

    def decrypt(self, ciphertext: tuple) -> int:
        """
        Decrypt ciphertext with Elliptic Curve ElGamal
        Args:
            ciphertext (tuple): c1 and c2
        Returns:
            plaintext (int): restored message
        """
        # modulo
        p = self.curve.p

        # private key
        ka = self.keys["private_key"]["ka"]

        c1, c2 = ciphertext
        c1_prime = (c1[0], (-1 * c1[1]) % p)
        s_prime = self.curve.apply_double_and_add_method(G=c1_prime, k=ka, p=p)
        s_prime = self.curve.add_points(P=c2, Q=s_prime, p=p)

        # s_prime is a point on the elliptic curve
        # s_prime = k x G
        # we need to find k from known s_prime and G
        # this requires to solve ECDLP

        # base point
        G = self.curve.G
        k = 2
        while True:
            G = self.curve.add_points(P=G, Q=self.curve.G, p=p)
            if G[0] == s_prime[0] and G[1] == s_prime[1]:
                return k
            k = k + 1
            if k > self.curve.n:
                raise ValueError(f"Cannot restore scalar from {s_prime} = k x {self.curve.G}")

    def multiply(self, ciphertext1: tuple, ciphertext2: tuple) -> tuple:
        raise ValueError(
            "Elliptic Curve ElGamal is not homomorphic with respect to the multiplication"
        )

    def add(self, ciphertext1: tuple, ciphertext2: tuple) -> tuple:
        """
        Perform homomorphic addition on encrypted data
        Result of this must be equal to E(m1 + m2)
        Args:
            ciphertext1 (dict): Elliptic Curve ElGamal ciphertext consisting of c1 and c2 keys
            ciphertext2 (dict): Elliptic Curve ElGamal ciphertext consisting of c1 and c2 keys
        Returns
            ciphertext (dict): Elliptic Curve ElGamal ciphertext consisting of c1 and c2 keys
        """
        a = self.curve.add_points(P=ciphertext1[0], Q=ciphertext2[0], p=self.curve.p)
        b = self.curve.add_points(P=ciphertext1[1], Q=ciphertext2[1], p=self.curve.p)
        return a, b

    def xor(self, ciphertext1: tuple, ciphertext2: tuple) -> int:
        raise ValueError(
            "Elliptic Curve ElGamal is not homomorphic with respect to the exclusive or"
        )

    def multiply_by_contant(self, ciphertext: tuple, constant: int) -> tuple:
        """
        Multiply a ciphertext with a plain constant.
        Result of this must be equal to k x E(m1) = E(m1 * k)
        where E(m1) = ciphertext
        Args:
            ciphertext (int): ciphertext created with Elliptic Curve ElGamal
            constant (int): known plain constant
        Returns:
            ciphertext (int): new ciphertext created with Elliptic Curve ElGamal
        """
        return self.curve.apply_double_and_add_method(
            G=ciphertext[0], k=constant, p=self.curve.p
        ), self.curve.apply_double_and_add_method(G=ciphertext[1], k=constant, p=self.curve.p)

    def reencrypt(self, ciphertext: tuple) -> tuple:
        raise ValueError("Elliptic Curve ElGamal does not support regeneration of ciphertext")
